/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2009, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Peers facility.
 *
 * This file contains the code of the thread(s) that will attempt connection to a remote peer.
 *
 */

#include "waaad-internal.h"
#include "peer-internal.h"

static void * _peer_client_th(void * arg) 
{
	int ret = 0;
	int sock = 0;
	_peer_t * peer;
	uint16_t in = 0, out = 0;
	peer_add_t * infos;
	
	THREAD_NAME_PEER( "Connecting", arg );
	TRACE_ENTRY("%p", arg);
	
	/* Check the parameter */
	CHECK_PARAMS_DO( VALIDATE_PEER(arg), goto end );
	
	peer  = _P(arg);
	infos = &peer->p_addinfo;
	
	/* Now attempt the connection... we'll succeed or be canceled */
	switch (infos->pa_proto) {
#ifndef DISABLE_SCTP
		case IPPROTO_SCTP:
			out = infos->pa_streams;
			ret = _peer_sctp_connect( (sSA *) &infos->pa_ss, sSSlen(&infos->pa_ss), &sock, &out, &in );
			break;
#endif /* DISABLE_SCTP */
			
		case IPPROTO_TCP:
			ret = _peer_tcp_connect( (sSA *) &infos->pa_ss, sSSlen(&infos->pa_ss), &sock );
			break;
			
		default:
			CHECK_PARAMS_DO( 0, goto end );
	}
	
	if (ret == 0) {
		_ev_connected_t * ev;
		
		CHECK_MALLOC_DO( ev = (_ev_connected_t *) malloc(sizeof(_ev_connected_t)), goto error );
		memset(ev, 0, sizeof(_ev_connected_t));
		
		ev->proto = infos->pa_proto;
		ev->sock = sock;
		ev->ostr = out;
		ev->istr = in;
		
		CHECK_FCT_DO( _peer_events_send(peer, PEVENT_CONNECTED, ev), { free(ev); goto error; } );
		TRACE_DEBUG(INFO, "Connection to peer '%s' established", peer->p_diamid);
	}
		
end:
	return NULL;

error:
	CHECK_SYS_DO( shutdown(sock, SHUT_RDWR), /* continue */ );
	return NULL;
}

/* Start the connect thread(s) for a peer */
int _peer_client_start(_peer_t * peer)
{
	CHECK_PARAMS( VALIDATE_PEER(peer) );
	CHECK_PARAMS( peer->p_con_th == (pthread_t)NULL );
	
	/* Ok the peer is in appropriate state, let's start the thread */
	CHECK_POSIX(  pthread_create( &peer->p_con_th, NULL, _peer_client_th, peer )  );
	
	/* done */
	return 0;
}

/* Stop this receiving thread */
int _peer_client_stop(_peer_t * peer)
{
	CHECK_PARAMS( VALIDATE_PEER(peer) );
	
	if (peer->p_con_th != (pthread_t)NULL)
		return _thread_term(&peer->p_con_th);
	else
		return 0;
}

