/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2009, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Peers management module - internal header.
 * 
 *  This header contains the private definitions to the peer module. 
 * It should not be included outside this module.
 *
 *  The namespace for internal peers objects is _peer_*
 *
 *  Here is a little more design details:
 *
 *  Each peer to which the local host is directly connected is represented by a _peer_t object.
 * Such objects are created either explicitely by calls to peer_add, or implicitely when a new
 * incoming connection is accepted.
 *
 *  Once the peer_start function has been called, a thread is associated to each peer object to
 * handle this peer state machine, with the exception of peers in the STATE_DISABLED state.
 *
 *  In addition, the peer state machine thread creates more threads for the peer for specific tasks:
 * waiting for incoming messages on the socket, sending messages on the socket, ...
 *
 */

/* About the files making this module:
 *
 * peer-struct.c:
 *   contains the functions to manipulate the _peer_t structure.
 *
 * peer-tcp.c:
 * peer-sctp.c:
 *   contains the functions to closely related to  these protocols
 *
 * peer-cer-cea.c:
 *   contains the code to create and parse Diameter Base messages.
 * 
 * peer_send.c:
 * peer_receive.c:
 *   contains the code of the threads that are responsibles for sending and receiving data.
 *
 * peer-sec-ini.c:
 *   contains the code of a default security handler to be used before remote identity is known.
 *
 * peer-events.c:
 *   contains the code related to events for the peers.
 *
 * peer-expire.c:
 *   contains the expiry mechanism.
 *
 * peer-server.c:
 *   contains the code of the server threads that handle incoming connections.
 *
 * peer-psm.c:
 *   contains the code of the peer state machine.
 */

#ifndef _PEER_INTERNAL_H
#define _PEER_INTERNAL_H

#include "waaad-internal.h"

#include <sys/socket.h>
#include <ctype.h>

/***********************************************************************************/

/* Size of the peer hash table */
#ifndef _PEER_HASH_SIZE
#define _PEER_HASH_SIZE	8	/* between 0 and 31 */
#endif /* _PEER_HASH_SIZE */

/* Maximum size of a Diameter message that we accept to receive */
#ifndef DIAMETER_MSG_SIZE_MAX
#define DIAMETER_MSG_SIZE_MAX	65535	/* in bytes */
#endif /* DIAMETER_MSG_SIZE_MAX */

/* The value to set into Vendor-Id AVP when sending CER / CEA */
#ifndef MY_VENDOR_ID
#define MY_VENDOR_ID	0 	/* Reserved value to tell it must be ignored */
#endif /* MY_VENDOR_ID */


/* Timeout for establishing a connection */
#ifndef CNX_TIMEOUT
#define  CNX_TIMEOUT	10	/* in seconds */
#endif /* CNX_TIMEOUT */

/* Timeout for receiving a CER after incoming connection is established */
#ifndef INCNX_TIMEOUT
#define  INCNX_TIMEOUT	 20	/* in seconds */
#endif /* INCNX_TIMEOUT */

/* Timeout for receiving a CEA after CER is sent */
#ifndef CEA_TIMEOUT
#define  CEA_TIMEOUT	10	/* in seconds */
#endif /* CEA_TIMEOUT */

/* The timeout value to wait for answer to a DPR */
#ifndef DPR_TIMEOUT
#define DPR_TIMEOUT 	15	/* in seconds */
#endif /* DPR_TIMEOUT */

/* The timeout value to simulate a "no timeout" mode when peer is disabled */
#ifndef NO_TIMEOUT
#define NO_TIMEOUT 	(60*60*24*365) /* in seconds */
#endif /* NO_TIMEOUT */

/***********************************************************************************/

/* States of a peer */
typedef enum {
	/* Stable states */
	STATE_DISABLED = 1,	/* No connexion must be attempted / only this state means that the peer PSM thread is not running */
	STATE_OPEN,		/* Connexion established */
	
	/* Peer state machine */
	STATE_CLOSED,		/* No connection established, will re-attempt after TcTimer. */
	STATE_CLOSING,		/* the connection is being shutdown (DPR/DPA in progress) */
	STATE_WAITCNXACK,	/* Attempting to establish transport-level connection */
	STATE_WAITCNXACK_ELEC,	/* Received a CER from this same peer on an incoming connection (other peer object), while we were waiting for cnx ack */
	STATE_WAITCEA,		/* Connection established, CER sent, waiting for CEA */
	/* STATE_WAITRETURNS_ELEC, */	/* This state is not stable and therefore deprecated:
				   We have sent a CER on our initiated connection, and received a CER from the remote peer on another connection. Election.
				   If we win the election, we must disconnect the initiated connection and send a CEA on the other => we go to OPEN state.
				   If we lose, we disconnect the other connection (receiver) and fallback to WAITCEA state. */
	
	/* Failover state machine */
	STATE_SUSPECT,		/* A DWR was sent and not answered within TwTime. Failover in progress. */
	STATE_REOPEN		/* Connection has been re-established, waiting for 3 DWR/DWA exchanges before putting back to service */
	
} _peer_state_t;

/* Convert a state to a string */
extern char * statestr[];
#define STATESTR_DEFINITION	\
char * statestr[] = {		\
	"<INVALID>",		\
				\
	"DISABLED", 		\
	"OPEN",			\
				\
	"CLOSED",		\
	"CLOSING",		\
	"WAIT_CNX_ACK",		\
	"WAIT_CNX_ACK_ELEC",	\
	"WAIT_CEA",		\
				\
	"SUSPECT",		\
	"REOPEN"		\
}			
#define VALSTATE( _state_ ) ( (((_state_) > 0) && ((_state_) <= STATE_REOPEN)) ? (_state_) : 0 )
#define STATESTR( _st_ ) statestr[VALSTATE( _st_ )]

struct _ev_incoming;

/***********************************************************************************/


/* Flags definitions */
#define PEERFL_DW_PENDING		( 1 << 0 )	/* A DWR message was sent and not answered yet */
#define PEERFL_CNX_PB			( 1 << 1 )	/* The peer was disconnected because of watchdogs; must exchange 3 watchdogs before putting back to normal */
#define PEERFL_DW_COUNTER_Lo		( 1 << 2 )	/* mask for DW counter after a peer connection is re-established after an interruption (low bit) */
#define PEERFL_DW_COUNTER_Hi		( 1 << 3 )	/* mask for DW counter after a peer connection is re-established after an interruption (high bit) */
#define PEERFL_SHUTTING_DOWN		( 1 << 4 )	/* A DPR request was sent */
#define PEERFL_DISABLE_AFTER_SHUTDOWN	( 1 << 5 )	/* When the peer enters the "CLOSED" state, its PSM must be stopped */
#define PEERFL_DESTROY_AFTER_SHUTDOWN	( 1 << 6 )	/* When the peer enters the "CLOSED" state, its PSM must be stopped and resources freed */
#define PEERFL_RESPONDER             	( 1 << 7 )	/* The peer is a responder */


/* Peer internal description */
typedef struct {
	/* For debug */
	uint32_t	 p_eyec;	/* An eyecatcher to verify object is valid. Must be PEER_EYEC. */
	
	/* Chaining of the peer in global lists */
	uint32_t	 p_hash;	/* Hash value of the p_diamid */
	uti_list_t	 p_global;	/* Hash table sublist this peer belongs to */

	/* Peer data */
	char 		*p_diamid;	/* The Diameter-Id of this peer, once known. All lower-case. */
	char 		*p_realm;	/* The realm of the peer. */
	uint32_t	 p_hbh;		/* next Hop-by-hop free value for sending messages */
	
	/* Peer state */
	uint32_t	 p_flags;	/* The PEERFL_* flags */
	_peer_state_t	 p_state;	/* State of the peer */
	uti_list_t	 p_active;	/* Hash sublist containing only the active peers (in STATE_OPEN) */
	pthread_t	 p_psm;		/* The thread handling this peer state machine. */
	pthread_mutex_t	 p_lock;	/* Mutex to protect this object */
	pthread_cond_t	 p_condvar;	/* The cond var to be waited in the peer state machine (for timeouts and DWR and events) */
	struct timespec	 p_ts;		/* multi-purpose timespec (meaning depends on the state) managed by the peer state machine */
	uti_list_t	 p_events;	/* queue of events (_pe_t) received by the peer */
	
	/* Messages handling */
	meq_t 		*p_in_q;	/* Incoming (received) message queue */
	pthread_t	 p_in_th;	/* The thread handling messages reception, created by p_psm */
	meq_t 		*p_out_q;	/* Outgoing message queue */
	pthread_t	 p_out_th;	/* The thread handling messages envoy, created by p_psm */
	uti_list_t	 p_sent;	/* List of sent requests without answer. Items are p_sent_t structures. */
	
	/* Initial information and expiration list */
	peer_add_t	 p_addinfo;	/* Initial informations received to connect to the peer */
	uti_list_t	 p_exp_list;	/* list of expiring threads, sorted by expiration date */
	
	/* Connection information */
	int		 p_sock;	/* We use standard Berkeley sockets for both TCP and SCTP connections */
	uint16_t	 p_ostr;	/* number of outbound streams (for SCTP) */
	uint16_t	 p_curstr;	/* id of the last used out stream */
	uint16_t	 p_istr;	/* number of inbound streams (for SCTP) */
	pthread_t	 p_con_th;	/* The thread handling initial connection */
	size_t		 p_peeraddr_sz;	/* Number of items in the array p_peeraddr bellow */
	sSS 		*p_peeraddr;	/* Array of the attachment points of the remote peer (received in Host-IP-Address AVPs) */
	
	/* Security information */
	uti_list_t 	 p_sec_list;	/* Used to store possible security mechanisms for this peer, ordered by priority. elements are _sec_item_t structures. */
	sec_mod_hdl_t 	*p_sec_hdl;	/* store reference to the security handler, to call sec_modunlink later */
	sec_module_t	*p_secmod;	/* the security callbacks */
	sec_session_t	 p_sec_session;	/* structure passed back to the security module, for connection information */
	void		*p_ext_session;	/* opaque data that can be used by the security extension to store internal states */
	size_t		 p_remotesec_sz;/* number of elements in the next array */
	uint32_t	*p_remotesec;	/* array of security mechanisms requested by remote peer in a CER (for responder) */
	
	/* Supported applications (result of CER/CEA) -- contains all applications advertized by remote */
	size_t		 p_app_size;	/* The size of the array pointed by p_app_list */
	peer_appl_t	*p_app_list;	/* points to an array of supported applications, of size p_app_size + 1, and the last element is always 0 */
	int		 p_app_relay;	/* Did the remote host advertize the relay application? */
	
	/* Information about the remote peer software */
	vendor_id_t	 p_vendor;	/* Content of the Vendor-Id AVP or 0 if undefined */
	char		*p_prodname;	/* Product name, 0-terminated UTF8 string */
	uint32_t	 p_firmrev;	/* The firmware revision */
	uint32_t	 p_orstate;	/* The origin-state id of the remote peer. May be used to detect resets of the remote peer (?) */
	
	/* Saved data that will be used in case of election */
	struct _ev_incoming *p_electdata;	/* The data associated with the incoming CER */
	
} _peer_t;

/* The eye-catcher value */
#define PEER_EYEC	0x023350B7

/* Cast macro */
#define _P( _peerptr_ ) ((_peer_t *)( _peerptr_ ))

/* Check macro */
#define VALIDATE_PEER( _peer_ ) 				\
	(((_peer_) != NULL) 					\
	&& (_P(_peer_)->p_eyec == PEER_EYEC) 			\
	&& (_P(_peer_)->p_exp_list.head != &_peers_deleted))	


/* The peers hash table. ALL peers are linked in the "all" hash list, by their p_global list */
typedef struct {
	uti_list_t	all;	/* Sentinel for the p_global list of peers which hash belongs to this sublist. */
	pthread_mutex_t	lock;	/* to protect these lists */
	int		count;	/* number of peers in this sublist */
} _peer_hash_t;

extern _peer_hash_t _peer_hash[1 << _PEER_HASH_SIZE];

#define H_MASK( __hash) ((__hash) & (( 1 << _PEER_HASH_SIZE ) - 1))
#define H_ALL(  _hash ) (&(_peer_hash[H_MASK(_hash)].all    ))
#define H_LOCK( _hash ) (&(_peer_hash[H_MASK(_hash)].lock   ))
#define H_COUNT(_hash ) (  _peer_hash[H_MASK(_hash)].count   )

/* the others lists of peers */
extern uti_list_t	_peers_actives;	  /* peers in the OPEN state, linked by their p_active fields. */
extern uti_list_t 	_peers_deleted;   /* deleted peers, threads are linked by their p_exp_list list. */
extern uti_list_t 	_peers_expiring;  /* peers that will expire, linked by their p_exp_list list. */
extern pthread_mutex_t  _peers_lists_mtx; /* mutex to protect these lists */

#define THREAD_NAME_PEER( name, peer )	{				\
	_peer_t * __p = _P(peer);					\
	log_set_thread_name( (name), __p?__p->p_diamid:"no id" );	\
}

/***********************************************************************************/

/* Events that can be sent to the peer and handled by the peer state machine */
typedef enum {
	PEVENT_SHUTDOWN = 1,	/* The daemon requested to initiate shutdown of this peer (will go to disabled state) */
	PEVENT_DESTROY,		/* The daemon requested to destroy this peer (any resources will be freed) */
	PEVENT_EXPIRE,		/* the peer lifetime is expiring; call p_addinfo.pa_cb or delete */
			
	PEVENT_TIMEOUT,		/* the timeout for this state has expired */
	
	PEVENT_CONNECTED,	/* the socket has been connected */
	PEVENT_MSGRCVD,		/* a message was received */
	PEVENT_INCNX,		/* Incoming connection from this peer */
	
	PEVENT_ASSOC_CHG,	/* association changed (SCTP notification) */
	PEVENT_PEERADDR_CHG,	/* peer address changed (SCTP notification) */
	PEVENT_REMOTE_ERR,	/* Remote error (SCTP notification) */
	
	PEVENT_DISCONNECTED,	/* the socket has been disconnected */
	PEVENT_SND_FAILED,	/* Failed to send a message (means probably connection is broken) */
	PEVENT_TH_TERM_IN,	/* The incoming thread terminated (on error) */
	PEVENT_TH_TERM_OUT,	/* The outgoing thread terminated (on error) */
	
	PEVENT_MAX	/* Invalid value */
} _pevent_t;

/* Convert an event to a string */
extern char * eventstr[];
#define EVENTSTR_DEFINITION	\
char * eventstr[] = {		\
	"<INVALID>",		\
				\
	"SHUTDOWN", 		\
	"DESTROY", 		\
	"EXPIRE",		\
				\
	"TIMEOUT",		\
				\
	"CONNECTED",		\
	"MSGRCVD",		\
	"INCNX",		\
				\
	"ASSOC_CHG",		\
	"PEERADDR_CHG",		\
	"REMOTE_ERR",		\
				\
	"DISCONNECTED",		\
	"SND_FAILED",		\
	"TH_TERM_IN",		\
	"TH_TERM_OUT"		\
}
#define VALEVENT( _event_ ) ( (((_event_) > 0) && ((_event_) < PEVENT_MAX)) ? (_event_) : 0 )
#define EVENTSTR( _ev_ ) eventstr[VALEVENT( _ev_ )]

/* This type represents an event */
typedef struct {
	uti_list_t	list;	/* list of events; the "o" field is the event data, if appropriate */
	_pevent_t	event;	/* identifier of this event */
} _pe_t;

/* Structure for Incoming calls event data */
typedef struct _ev_incoming {
	_peer_t	* peer; /* The dummy peer containing all the CER data and incoming connection information */
	uint32_t  ete; /* the end-to-end id of the received CER to create the matching answer */
	uint32_t  hbh; /* the hop-by-hop id of the received CER to create the matching answer */
} _ev_incoming_t;

/* Structure for Connected event data */
typedef struct {
	int	 proto;	/* the protocol on which the connection was established, IPPROTO_TCP or IPPROTO_SCTP */
	int 	 sock;	/* the socket identifier of the new connection */
	uint16_t ostr;	/* # of out streams for SCTP */
	uint16_t istr;	/* # of in streams for SCTP */
} _ev_connected_t;

/***********************************************************************************/

/* Reference to the "no sec" security module used before CER is exchanged */
extern sec_module_t _peer_secini_mod;


/* Initialize a sec_session_t with callbacks for TCP or SCTP */
#define SEC_SESS_INIT( _sessobj_, _proto_, _peer_, _psock_ ) {	\
	sec_session_t * __sess = (sec_session_t *) (_sessobj_);	\
	peer_t * __peer = (peer_t *) (_peer_);			\
	sec_conn_t * __conn = (sec_conn_t *) (_psock_);		\
	memset(__sess, 0, sizeof(sec_session_t));		\
	__sess->peer = __peer;					\
	__sess->conn = __conn;					\
	__sess->proto = (_proto_);				\
	switch (__sess->proto) {				\
	case IPPROTO_TCP:					\
		__sess->cbs.tcp.send_data = 			\
			(sec_send_tcp_cb) _peer_tcp_send;	\
		__sess->cbs.tcp.recv_data = 			\
			(sec_recv_tcp_cb) _peer_tcp_recv;	\
		break;						\
	case IPPROTO_SCTP:					\
		__sess->cbs.sctp.send_data = 			\
			(sec_send_sctp_cb) _peer_sctp_send;	\
		__sess->cbs.sctp.recv_data = 			\
			(sec_recv_sctp_cb) _peer_sctp_recv;	\
		break;						\
	default: assert(0);					\
	}							\
}

/***********************************************************************************/
/* Store sent requests */
typedef struct {
	uti_list_t	chain;	/* linking. the "o" field is the msg_t*, which hbh is the *previous* one */
	uint32_t	hbh;	/* the hbh with which the message was sent */
} p_sent_t;

/***********************************************************************************/

/* Functions */

int _peer_struct_malloc(char * id, _peer_t **new);
int _peer_struct_unlink(_peer_t * peer);
int _peer_struct_destroy(_peer_t ** peer);
int _peer_struct_init();
int _peer_struct_fini();
int _peer_struct_search(char * diamid, size_t diamidlen, uint32_t * hash, uti_list_t ** peerglob);
int  _peer_struct_appl_add( _peer_t * peer, application_id_t a, vendor_id_t v, peer_app_type_t t);
void _peer_struct_appl_reset( _peer_t * peer );
int  _peer_struct_secid_add( _peer_t * peer, uint32_t secid);
void _peer_struct_secid_reset( _peer_t * peer );
int _peer_struct_secid_check( _peer_t * peer, uint32_t secid);
void _peer_struct_addrs_reset( _peer_t * peer );
void _peer_struct_secmod_reset(_peer_t * peer);
int peer_dump_hash(int debug_level);
void cleanup_peer(void * peer);


int _peer_failover(_peer_t * peer, int freeloc);

int _peer_expire_queue(_peer_t * peer);
int _peer_expire_init(void);
int _peer_expire_fini(void);

int _peer_events_send_nl(_peer_t * peer, _pevent_t event, void * data);
int _peer_events_send(_peer_t * peer, _pevent_t event, void * data);
void _peer_events_dispose_data( _pevent_t event, void ** data );
void _peer_events_reset( _peer_t * peer );

int _peer_in_start(_peer_t * peer);
int _peer_in_stop(_peer_t * peer);

int _peer_out_start(_peer_t * peer);
int _peer_out_stop(_peer_t * peer);
int _peer_out_sendmsg(_peer_t * peer, msg_t * msg);

#ifndef DISABLE_SCTP

int _peer_sctp_create_server( int * sock );
int _peer_sctp_accept( int sock, int * new_sd, uint16_t *ostreams, uint16_t *istreams );
int _peer_sctp_connect( const sSA *addr, socklen_t addrlen, int *sock, uint16_t *ostreams, uint16_t *istreams );
ssize_t _peer_sctp_send(sec_session_t * sess, int *streamid, const void * data, size_t len);
ssize_t _peer_sctp_recv(sec_session_t * sess, uint16_t * streamid, void ** data, size_t *len);

#else /* DISABLE_SCTP */

# ifndef IPPROTO_SCTP
#  define IPPROTO_SCTP 132
# endif /* IPPROTO_SCTP */

 static inline ssize_t _peer_sctp_send(sec_session_t * sess, int *streamid, const void * data, size_t len)
 {
	TRACE_HERE();
	return -1;
 }
 static inline ssize_t _peer_sctp_recv(sec_session_t * sess, uint16_t * streamid, void ** data, size_t *len)
 {
	TRACE_HERE();
	return -1;
 }
#endif /* DISABLE_SCTP */

int _peer_tcp_create_server( int * sock, sSA *addr, socklen_t salen );
int _peer_tcp_connect( const sSA *addr, socklen_t addrlen, int *sock );
ssize_t _peer_tcp_send( sec_session_t * sess, void *data, size_t len);
ssize_t _peer_tcp_recv( sec_session_t * sess, void *data, size_t len);

int _peer_cer_cea_init(void);
int _peer_cer_cea_fini(void);
int _peer_cer_create( _peer_t * peer, msg_t ** cer );
int _peer_cer_parse( msg_t * cer, _peer_t * peer );
int _peer_cea_create( _peer_t * peer, msg_t ** cea, uint32_t eteid, uint32_t hbhid );
int _peer_cea_parse( msg_t * cea, _peer_t * peer );

int _peer_dwr_dwa_init(void);
int _peer_dwr_dwa_fini(void);
int _peer_dwr_create( _peer_t * peer, msg_t ** dwr );
int _peer_dwr_parse( msg_t * dwr, _peer_t * peer );
int _peer_dwa_create( _peer_t * peer, msg_t ** dwa, uint32_t eteid, uint32_t hbhid );
int _peer_dwa_parse( msg_t * dwa, _peer_t * peer );

int _peer_dpr_dpa_init(void);
int _peer_dpr_dpa_fini(void);
int _peer_dpr_create( _peer_t * peer, msg_t ** dpr, uint32_t cause );
int _peer_dpr_parse( msg_t * dpr, _peer_t * peer );
int _peer_dpa_create( _peer_t * peer, msg_t ** dpa, uint32_t eteid, uint32_t hbhid );
int _peer_dpa_parse( msg_t * dpa, _peer_t * peer );

int _peer_server_start();
int _peer_server_stop();

int _peer_client_start(_peer_t * peer);
int _peer_client_stop(_peer_t * peer);

int _peer_psm_start(_peer_t * peer, int mode);
int _peer_psm_start_all(void);
int _peer_psm_stop(_peer_t * peer, int mode);
#ifdef IN_WAAAD_TEST
void _psm_change_state(_peer_t * peer, _peer_state_t new);
#endif /* IN_WAAAD_TEST */


#endif /* ! _PEER_INTERNAL_H */

