/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2009, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

#include "tests.h"

/* Test for the dict_iterate_rules function */
int iter_test(void * data, dict_rule_data_t * rule)
{
	dict_avp_data_t avpdata;
	(*(int *)data)++;
	
	CHECK( 0, dict_getval ( rule->rule_avp, &avpdata ) );
	TRACE_DEBUG(FULL, "rule #%d: avp '%s'", *(int *)data, avpdata.avp_name);
	return 0;
}

/* Main test routine */
int main(int argc, char *argv[])
{
	/* First, initialize the daemon modules */
	INIT_WAAAD();
	
	/* Test creating and searching all types of objects */
	{
		dict_object_type_t type;
		dict_object_t * obj1 = NULL;
		dict_object_t * obj2 = NULL;
		dict_object_t * obj3 = NULL;

		vendor_id_t vendor_id = 735671;
		dict_vendor_data_t vendor1_data = { 735671, "Vendor test 1" };
		dict_vendor_data_t vendor2_data = { 735672, "Vendor test 2" };
		dict_application_data_t app1_data = { 735674, "Application test 1" };
		
		
		/* Create two vendors */
		CHECK( 0, dict_new ( DICT_VENDOR, &vendor1_data , NULL, &obj1 ) );
		CHECK( 0, dict_new ( DICT_VENDOR, &vendor2_data , NULL, NULL ) );
		
		/* Check we always retrieve the correct vendor object */
		CHECK( 0, dict_search ( DICT_VENDOR, VENDOR_BY_ID_REF, &vendor_id, &obj2 ) );
		CHECK( obj1, obj2);
		CHECK( 0, dict_search ( DICT_VENDOR, VENDOR_BY_ID_VAL, (void *)(unsigned long)vendor_id, &obj2 ) );
		CHECK( obj1, obj2);
		CHECK( 0, dict_search ( DICT_VENDOR, VENDOR_BY_NAME, "Vendor test 1", &obj2 ) );
		CHECK( obj1, obj2);
		
		/* Check the error conditions */
		CHECK( 0, dict_search ( DICT_VENDOR, VENDOR_BY_ID_REF, &vendor_id, NULL ) );
		
		vendor_id = 735673; /* Not defined */
		CHECK( ENOENT, dict_search ( DICT_VENDOR, VENDOR_BY_ID_REF, &vendor_id, NULL ) );
		CHECK( ENOENT, dict_search ( DICT_VENDOR, VENDOR_BY_ID_VAL, (void *)(unsigned long)vendor_id, NULL ) );
		CHECK( ENOENT, dict_search ( DICT_VENDOR, VENDOR_BY_NAME, "Vendor test 3", NULL ) );
		
		/* Check the get_* functions */
		CHECK( 0, dict_gettype ( obj1, &type ) );
		CHECK( DICT_VENDOR, type );
		CHECK( 0, dict_getval ( obj1, &vendor1_data ) );
		CHECK( 735671, vendor1_data.vendor_id );
		CHECK( 0, strcmp(vendor1_data.vendor_name, "Vendor test 1") );
		/* error conditions */
		CHECK( EINVAL, dict_gettype ( "not an object", &type ) );
		CHECK( EINVAL, dict_getval ( "not an object", &vendor1_data ) );
		
		/* Create the application with vendor1 as parent */
		CHECK( EINVAL, dict_new ( DICT_APPLICATION, &app1_data , "bad object", &obj2 ) );
		CHECK( 0, dict_new ( DICT_APPLICATION, &app1_data , obj1, &obj2 ) );
		
		CHECK( 0, dict_search ( DICT_VENDOR, VENDOR_OF_APPLICATION, obj2, &obj3 ) );
		CHECK( obj1, obj3);
		
		/* Creating and searching the other objects is already done in dictionary initialization */
		
	}

	/* Test creation of the "Example-AVP" grouped AVP from the RFC */
	{
		int nbr = 0;
		dict_object_t * origin_host_avp = NULL;
		dict_object_t * session_id_avp = NULL;
		dict_object_t * example_avp_avp = NULL;
		dict_rule_data_t rule_data = { NULL, RULE_REQUIRED, -1, -1, 1 };
		dict_avp_data_t example_avp_data = { 999999, 0, "Example-AVP", AVP_FLAG_VENDOR , 0, AVP_TYPE_GROUPED };

		CHECK( 0, dict_search ( DICT_AVP, AVP_BY_NAME, "Origin-Host", &origin_host_avp ) );
		CHECK( 0, dict_search ( DICT_AVP, AVP_BY_NAME, "Session-Id", &session_id_avp ) );
		
		CHECK( 0, dict_new ( DICT_AVP, &example_avp_data , NULL, &example_avp_avp ) );
		
		rule_data.rule_avp = origin_host_avp;
		rule_data.rule_min = 1;
		rule_data.rule_max = 1;
		CHECK( 0, dict_new ( DICT_RULE, &rule_data, example_avp_avp, NULL ) );
		
		rule_data.rule_avp = session_id_avp;
		rule_data.rule_min = 1;
		rule_data.rule_max = -1;
		CHECK( 0, dict_new ( DICT_RULE, &rule_data, example_avp_avp, NULL ) );
		
		CHECK( 0, dict_iterate_rules ( example_avp_avp, &nbr, iter_test) );
		CHECK( 2, nbr );
	}

	/* That's all for the tests yet */
	PASSTEST();
} 
	
